#    CamPlay (C) 2011, released under the GPL 3.0
#    
#    This file is part of the CamPlay Camera Game Library.
#
#    CamPlay is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    CamPlay is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with CamPlay.  If not, see <http://www.gnu.org/licenses/>.


import pygame
from pygame.color import THECOLORS
import pygame.camera
import math, sys, random
import time
from pygame.locals import *
from imagecache import *
import ui


__author__ = "Michael Bethencourt"
__copyright__ = "Copyright 2011, Michael Bethencourt"
__license__ = "GPL"
__version__ = "3.0.9"
__email__ = "michaelpb@gmail.com"
__version__ = "0.1"


pygame.init()
pygame.camera.init()
WIDTH = 640 
HEIGHT = 480 

CAMERA_DIMENSIONS = (WIDTH, HEIGHT)


FATNESS = 20000
VAGUENESS = 40
CYCLE = 2

MOTION_MASK_FUZZINESS = 3
MATTER_MASK_FUZZINESS = 2

MOTION_LENGTH = 3

class CamPlay(object):
	MODE_2D, SINGLE, DOUBLE, MODE_3D, RIGHT, LEFT, DEMO = range(0,7)
	def __init__(self, display, size, 
					mode=MODE_2D, 
					debug=True, dont_start=False, vflip=True,
					clock=None, autopause=True, 
					):
		self.size = size
		# create a display surface. standard pygame stuff
		self.display = display
		self.debug = debug
		self.mode = mode
		self.vflip = vflip
		self.clock = clock
		self.autopause = autopause
		self.are_players_touching = False
		self.players = []
		self.player_sprite_group = pygame.sprite.Group()
		
		# this is the same as what we saw before
		clist = pygame.camera.list_cameras()
		if not clist:
			raise ValueError("Sorry, no cameras detected.")
		self.cam = pygame.camera.Camera(clist[0], CAMERA_DIMENSIONS)
		if not dont_start:
			self.start()
		
		# create a surface to capture to.  for performance purposes
		# bit depth is the same as that of the display surface.
		# TODO: fix issue with inconsistent dimensions --- it should just 
		# work seemlessly regardless of pygame dimensions
		self.snapshot = pygame.surface.Surface(self.size, 0, self.display)
		self.background = pygame.surface.Surface(self.size, 0, self.display)
		self.thresholded = pygame.surface.Surface(self.size, 0, self.display)

		self.old_snapshots = [self.snapshot]

		# Set up UI
		if clock:
			self.modal = ui.ModalUI(self, display, clock)
	
	def start(self):
		self.cam.start()

	def stop(self):
		self.cam.stop()
		
	def new_game(self, players, player_sprite=None, autopause=True):
		self.autopause = autopause
		if not player_sprite:
			player_sprite = PlayerSprite
		self.player_number = players
		if players==self.SINGLE:
			self.players = [Player(self.SINGLE, player_sprite, self)]
		elif players==self.DOUBLE:
			self.players = [Player(self.LEFT, player_sprite, self), 
						Player(self.RIGHT, player_sprite, self)]
		elif players==self.DEMO:
			self.palyers = []
		
		for player in self.players:
			self.player_sprite_group.add(player.sprite)


	def _get_camera_surface(self):
		self.snapshot =  self.cam.get_image(self.snapshot)
		self.snapshot = pygame.transform.flip(self.snapshot, True, False)
		if self.size != CAMERA_DIMENSIONS:
			print "NOT EQUAL"
			self.snapshot.scale(CAMERA_DIMENSIONS)
		return self.snapshot
	
	def calibrate(self):
		# capture a bunch of background images
		bg = []
		for i in range(0,5):
			bg.append(self._get_camera_surface().copy())
		# average them down to one to get rid of some noise
		pygame.transform.average_surfaces(bg, self.background)
		# blit it to the display surface
		self.display.blit(self.background, (0,0))
		pygame.display.flip()

	'''
	@staticmethod
	def blit_mask(image, mask, mask_rect):
		"""
		VERY SLOW, use sparingly.
		Blit an source image to the dest surface, at destpos, with a mask, using
		only the maskrect part of the mask.
		"""
		tmp = source.copy()
		tmp.blit(mask, maskrect.topleft, maskrect, special_flags=pygame.BLEND_RGBA_MULT)
		dest.blit(tmp, destpos, dest.get_rect().clip(maskrect))
	
	@staticmethod
	def make_surface_mask(target, source, background):
		# Threshold
		return target
	'''

	@staticmethod
	def make_mask(target, source, background):
		# Threshold
		pygame.transform.threshold(target, source,
				(0,255,0),(30,30,30),(0,0,0),1, background)
		#def blit_mask(source, dest, destpos, mask, maskrect):
		mask = pygame.mask.from_threshold(target, (0, 0, 0), (10, 10, 10, 10))
		return mask
	
	@staticmethod
	def dilate_mask(mask, amount=2):
		# scale down
		size = mask.get_size()
		new_size = (size[0]/amount, size[1]/amount)
		if not isinstance(mask, pygame.Surface):
			small = mask.scale(new_size)
			# scale back
			return small.scale(size)
		else:
			small = pygame.transform.scale(mask, new_size)
			return pygame.transform.scale(mask, size)

	def gen_matter_mask(self):
		mask = CamPlay.make_mask(self.thresholded, self.snapshot, self.background)
		mask = CamPlay.dilate_mask(mask, MATTER_MASK_FUZZINESS)
		return mask


	def get_motion_mask(self, snapshot, fuzziness=MOTION_MASK_FUZZINESS):
		last_snapshot = self.old_snapshots[0]
		motion_mask = CamPlay.make_mask(self.thresholded, last_snapshot, snapshot)
		motion_mask = CamPlay.dilate_mask(motion_mask, MOTION_MASK_FUZZINESS)
		self.old_snapshots.append(snapshot.copy())
		# TODO: compute multiple motion speeds :)
		if len(self.old_snapshots) >= MOTION_LENGTH:
			self.old_snapshots.pop(0)
		return motion_mask
	
	def tick(self, blit=False, detect_motion=True):
		self.snapshot = self._get_camera_surface()
		if blit:
			self.display.blit(self.snapshot, (0,0))
		mask = self.gen_matter_mask()
		
		# Do motion detection
		if detect_motion:
			motion_mask = self.get_motion_mask(self.snapshot)
		
		if self.player_number == self.SINGLE:
			self._find_player(mask, motion_mask)
		else:
			self._find_both_players(mask, motion_mask)

	def _find_player(self, mask, motion_mask):
		blob = mask.connected_component()
		if blob.count() < FATNESS/3:
			if self.autopause:
				pygame.mixer.music.pause()
				self.modal.get_in_frame_sequence(1) # wait for 2
				pygame.mixer.music.unpause()
			else:
				self.players[0].tick_offscreen()
		else:
			self._update_player(self.players[0], blob, motion_mask)
		
	def _find_both_players(self, mask, motion_mask):
		# Get the two blobs
		is_touching = False
		blobs = mask.connected_components(FATNESS)
		if len(blobs) < 1: 
			# TODO generate exception if no blocking wanted
			#print "No blobs!!!"
			if self.autopause:
				pygame.mixer.music.pause()
				self.modal.get_in_frame_sequence(2) # wait for 2
				pygame.mixer.music.unpause()
			else:
				self.player[0].tick_offscreen()
				self.player[1].tick_offscreen()
			return
		elif len(blobs) < 2:
			is_touching = True
			blob = blobs[0]
			height = blob.get_size()[1]
			x = blob.centroid()[0]
			for y in xrange(height):
				# Clear a horizontal line in the center
				blob.set_at((x, y), 0)
				blob.set_at((x-1, y), 0)
				blob.set_at((x+1, y), 0)
			blobs = blob.connected_components(FATNESS)
			tries = 0
			while len(blobs) < 2:
				#print "FATTNESS Descent mode, try number ", tries
				blobs = blob.connected_components(FATNESS/(tries+2))
				tries += 1
				# We give up soon, since its probably just one.
				if tries > 2:
					pygame.mixer.music.pause()
					self.modal.get_in_frame_sequence(2) # wait for 2
					pygame.mixer.music.unpause()
					return
			#print "One blob, got split successfully!"
		# sort by size
		blobs.sort(key=lambda b: b.count())
		# get the top two
		player_blobs = blobs[0:2]
		# separate into left and right
		player_blobs.sort(key=lambda b: b.centroid()[0])
		left_blob, right_blob = player_blobs
		# update players
		self._update_player(self.players[0], left_blob, \
					motion_mask, is_touching)
		self._update_player(self.players[1], right_blob, \
					motion_mask, is_touching)
		
	def _update_player(self, player, blob, motion_mask, is_touching=False):
		center = blob.centroid()
		if motion_mask:
			motion_mask = motion_mask.overlap_mask(blob, (0, 0))
		player.tick(center, blob, motion_mask, is_touching)
		if self.debug:
			points = blob.outline(VAGUENESS)
			if len(points)>2:
				pygame.draw.polygon(self.display, (255, 255, 0), points, 3)
			if motion_mask:
				points = motion_mask.outline(VAGUENESS)
				if len(points)>2:
					pygame.draw.polygon(self.display, (255, 0, 255), points, 5)

	def debug_outline_mask(self, display, mask):
		for blob in mask.connected_components(100):
			points = blob.outline(VAGUENESS)
			if len(points)>10:
				pygame.draw.polygon(display, (255, 255, 0), points, 3)


		
	def get_events(self):
		pass
		
	def show_menu(self, menu):
		pass

	def blit_ui(self, surf=None):
		if not surf:
			surf = self.display
		if len(self.players) == 1:
			txt = "%i" % self.players[0].score
			ui.blit_text(surf, txt, center=surf.get_rect().midtop)
		else:
			txt = "%i" % self.players[0].score
			ui.blit_text(surf, txt, left=surf.get_rect().topleft)
			txt = "%i" % self.players[1].score
			ui.blit_text(surf, txt, right=surf.get_rect().topright)

	def blit_hud(self, surf=None):
		if not surf:
			surf = self.display
		hud = ui.AutoGridContainer(*[p.hud for p in self.players])
		hud.render(surf, surf.get_rect())

class Player(CachedMutable):
	MODE_2D, SINGLE, DOUBLE, MODE_3D, RIGHT, LEFT = range(0,6)
	class Buff(object): 
		def __init__(self, name, image,  expire=None, max_count=None):
			self.name = name
			self.image = image
			self.expire = expire
			self.max_count = max_count
	def __init__(self, team, player_sprite, camplay):
		self.team = team
		self.camplay = camplay
		CachedMutable.__init__(self)
		self.info = {}
		self.center = (0,0)
		self.is_visible = False
		self.is_touching = False
		self.motion_mask = None
		self.offscreen_side = None
		# Add images to here to get them displayed as images
		# in the HUD
		self.buff_images = []
		self.buffs = []
		# Use this to keep track of score! 
		self.score = 0
		self.hud = ui.PlayerHUD(self)
		self.stats = {}
		self.sprite = player_sprite(self)
		self.mask = pygame.Mask((0,0))
	
	@property
	def name(self):
		if self.team == self.SINGLE:
			return "Player"
		return "Right player" if self.team==self.RIGHT else "Left player" 

	# REVAMP this shit --vvv
	def add_buff(self, name, image, expire=None, max_count=None):
		if max_count:
			existing_buffs = self.get_buffs(name)
			if len(existing_buffs) >= max_count:
				min_buff = min(self.buffs, key=lambda b: b.expire)
				self.buffs.remove(min_buff)
		self.buffs.append(Player.Buff(name, image, expire, max_count))

	
	def remove_buff(self, name):
		for buff in self.get_buffs(name):
			self.buffs.remove(buff)
	
	def get_buffs(self, name):
		return filter(lambda b: b.name==name, self.buffs)
	
	@mutable
	def tick(self, center, mask, motion_mask=None, \
				is_touching=False, surface_mask=None):
		#print "I AM IN TICK"
		self.is_visible = True
		#print "IS VISIBLE", self.is_visible
		self.offscreen_side = None
		self.center = center
		self.mask = mask
		self.motion_mask = motion_mask
		self.is_touching = is_touching
		self.surface_mask = surface_mask
		
		# It caches the mask and stuff
		self.sprite._reset_cache()
		
		self.tick_buffs()

		
	SIDE_LEFT, SIDE_TOP, SIDE_BOTTOM, \
		SIDE_RIGHT = 'left', 'top', 'bottom', 'right'
	@mutable
	def tick_offscreen(self):
		"""
		This is called when the person moves off screen
		"""
		self.mask = pygame.Mask((1,1))
		self.motion_mask = pygame.Mask((1,1))
		self.is_touching = False
		self.surface_mask = None
		
		self.is_visible = False
		self.tick_buffs()
		self.sprite._reset_cache()
		
		# If this is the first tick offscreen, we try to guess as to which
		# side we are at, and set things appropriately
		if self.offscreen_side is None:
			
			diffs = {
				self.SIDE_LEFT: self.center[0],
				self.SIDE_RIGHT: self.camplay.size[0] - self.center[0],
				self.SIDE_TOP: self.center[1],
				self.SIDE_BOTTOM: self.camplay.size[1] - self.center[1],
			}
			side, diff = min(diffs.items(), key=lambda a: a[1])
			self.center = self.center
			self.offscreen_side = side
			self.center = {
				self.SIDE_LEFT: (-1, self.center[1]),
				self.SIDE_RIGHT: (self.camplay.size[0]+1, self.center[1]),
				self.SIDE_TOP: (self.center[0], -1),
				self.SIDE_BOTTOM: (self.center[0], self.camplay.size[1]+1),
			}[side]
	
	def tick_buffs(self):
		"""
		Ticks "buffs", or power-ups, if the feature is in use.
		"""
		for buff in self.buffs:
			if buff.expire is not None:
				buff.expire -= 1
				if buff.expire <= 0:
					self.buffs.remove(buff)
	
	@property
	def rect(self):
		if self.is_visible:
			return self.mask.get_bounding_rects()[0]
		else:
			return pygame.Rect(0,0,0,0)
	
	@cached
	def get_size(self):
		return self.mask.count()

	@cached
	def get_center2d(self):
		if self.is_visible:
			return self.center
		else:
			return (-1, -1)
	
	@cached
	def get_center3d(self):
		# returns 3-tuple, of center with respect to space.  Distance is guessed by size
		pass

	@cached
	def get_motion_mask(self):
		if not self.motion_mask:
			return pygame.Mask((1,1))
			raise Exception("motion_mask not enabled")
		return self.motion_mask

	@cached
	def get_non_changing_motion_amount(self):
		# Used for WALK
		pass

	@property
	def is_touching_other_player(self):
		self.are_players_touching = False
		return self.is_touching

	@cached
	def get_speed(self):
		# Get speed of bounding box
		pass
	
	@cached
	def get_surface_of_player(self):
		# apply mask to player, to 'cut out'
		# cropped to bounding box
		# In colliding players during two player games, cut down the middle, like the mask, etc
		pass
	
	@cached
	def get_surface_of_player(self):
		pass

	@cached
	def listen_for(self, event_type, sensitivity):
		pass


class PlayerSprite(pygame.sprite.Sprite, CachedMutable):
	def __init__(self, player):
		self.player = player
		pygame.sprite.Sprite.__init__(self)
		CachedMutable.__init__(self)
	
	@property
	@cached
	def mask_surface(self):
		"""
		A BLACK & WHITE surface, representing the players mask
		(but as a surface, not a pygame mask). 
		"""
		raise NotImplemented()
		mask = self.mask.copy()
		
	@property
	@cached
	def image(self):
		# Currently not impleented :(
		return pygame.Surface((0,0), (0,0))
	
	@property
	@cached
	def mask(self):
		return self.player.mask

	@property
	@cached
	def motion_mask(self):
		return self.player.get_motion_mask()
	
	@property
	@cached
	def rect(self):
		rects = self.mask.get_bounding_rects()
		return rects[0] if rects else pygame.Rect((0,0), (0,0))



