import pygame
import os
import sys
import random
import math
import time
from pygame.locals import *
from imagecache import *
from ui import *

RECT_UNSET = pygame.Rect((-1, -1), (0, 0))

class Effect(dict):
	def __init__(self, *a, **k): 
		dict.__init__(self, *a, **k)
		for key, val in k.iteritems():
			setattr(self, key, val)
		self.period = k.get('period', None)
		self.blocking = k.get('blocking', False)
		self.after_finished = k.get('after_finished', None)
		self.easing = k.get('easing', lambda a, b: a)
		self.reset()

	def reset(self):
		self.tick = 0
		self.periodic = 0.0
		self.real_tick = 0
		self.real_periodic = 0.0
	
	def is_done(self):
		return False
	
	def update(self):
		self.real_tick += 1
		self.tick = self.easing(self.real_tick, self.period)
		if self.period:
			self.periodic_tick = (int(self.tick) % self.period)
			self.periodic = self.periodic_tick*1.0 / self.period

class EffectList(list):
	def __init__(self, *a, **k): 
		list.__init__(self, a)
		self.period = k.get('period', None)
		self.blocking = k.get('blocking', False)
		self.after_finished = k.get('after_finished', None)

	def is_done(self):
		return len(self) < 1

	def update(self):
		for effect in self:
			effect.update()

	def check_for_done(self, effect):
		if effect.is_done():
			# Effect is finished!  Replace our list with the action
			if effect.after_finished:
				self[:] = effect.after_finished(effect, self)
			self.remove_effect(effect)

	def remove_effect(self, effect):
		try:
			self.remove(effect)
		except ValueError: 
			# not in list
			pass
	
	def do(self, image, rect):
		image = image.copy()
		rect = rect.copy()
		for effect in self[:]:
			effect.update()
			image, rect = effect.do(image, rect)
			self.check_for_done(effect)
			if effect.blocking:
				# Effect is blocking, ie sequential, so skip
				# remaining effects
				break
		
		return image, rect
	
class effects:
	class Spin(Effect):
		def is_done(self):
			if 'end_angle' in self:
				print self.real_tick, self.period
				return self.real_tick >= self.period
			else:
				return False 
		def do(self, sprite, rect):
			start = self.get('start_angle', 0)*(180/3.14)
			end = self.get('end_angle', 6.28)*(180/3.14)
			angle = start + (self.periodic * (end - start))
			#print "ANGLE: ", angle
			new_sprite = pygame.transform.rotate(sprite, angle)
			new_rect = new_sprite.get_rect()
			new_rect.center = rect.center
			return new_sprite, new_rect

	class Glide(Effect): 
		def is_done(self):
			return self.real_tick >= self.period

		def do(self, image, rect):
			left_increment = (self.end[0] - self.start[0])*1.0 / self.period
			top_increment = (self.end[1] - self.start[1])*1.0 / self.period
			rect = rect.copy()
			rect.left = self.start[0]  + (left_increment * self.tick)
			rect.top = self.start[1]  + (top_increment * self.tick)
			return image, rect


	class Grow(Effect):
		def is_done(self):
			return self.real_tick >= self.period
		
		def do(self, image, rect):
			increment = (self.end_size - self.start_size)*1.0 / self.period
			new_size = self.start_size  + (increment * self.tick)
			if 'smooth' in self and self.smooth:
				scale = pygame.transform.smoothscale
			else:
				scale = pygame.transform.scale
			new_height = max(1, int(rect.height * new_size))
			new_width = max(1, int(rect.width * new_size))
			new_image = scale(image, (new_width, new_height))
			new_rect = new_image.get_rect()
			new_rect.center = rect.center
			return new_image, new_rect
	
	class FadeIn(Effect): 
		def is_done(self):
			return self.real_tick >= self.period
	
		def do(self, image, rect):
			alpha = self.periodic*255
			image.set_alpha(int(alpha), pygame.RLE_ACCEL)
			return image, rect
		
	class FadeOut(Effect): 
		"""COMPLETELY USELESS until I finish my image caching."""
		def is_done(self):
			return self.real_tick >= self.period
	
		def do(self, image, rect):
			print "ME!"
			alpha = (1.0-self.periodic)*255
			#image.set_alpha(int(alpha), pygame.RLEACCEL)
			image = change_alpha(image,alpha)
			print "ME!", alpha
			return image, rect

	
	class Explode(Effect): 
		pass
	
	class Kill(Effect): 
		def is_done(self):
			return True
		
		def do(self, image, rect):
			return image, RECT_UNSET
	
	class Loop(EffectList): 
		def reset(self):
			# TODO reset count
			EffectList.reset(self)
		def remove_effect(self, effect):
			# Have it append again if not done
			try:
				self.remove(effect)
			except ValueError: 
				# not in list
				pass
			if 'count' in self: 
				if self.count <= 0:
					# Last loop
					return
				self.count -= 1
			effect.reset()
			self.append(effect)

	class MusicSync(Effect):
		# A special blocking one, that "sort of blocks"
		pass
		
	@staticmethod
	def prune_all(effect, effect_list):
		return []
	
	@staticmethod
	def prune_before(effect, effect_list):
		ind = effect_list.index(effect)
		return effect_list[ind:]
	
	@staticmethod
	def prune_after(effect, effect_list):
		ind = effect_list.index(effect)
		return effect_list[:ind]

class easing:
	@staticmethod
	def _pow_easing(tick, period, exp):
		base = math.pow(period, 1.0/exp) * (tick*1.0/period)
		return int(math.pow(base, exp))

	@staticmethod
	def quad_slow(tick, period=100):
		# Quadratic easing
		return easing._pow_easing(tick, period, 2)
		# Is the following faster?
		#base = math.sqrt(period) * (tick*1.0/period)
		#return int(math.pow(base, 2))
	
	@staticmethod
	def cubic_slow(tick, period=100):
		# Cubic easing
		return easing._pow_easing(tick, period, 3)
	
	@staticmethod
	def quad_fast(tick, period=100):
		return period - easing.quad_slow(period-tick, period)
	
	@staticmethod
	def cubic_fast(tick, period=100):
		return period - easing.cubic_slow(period-tick, period)

	@staticmethod
	def cubic_slow_in_fast_out(tick, period=100):
		if tick<period/2:
			return int(easing.cubic_slow(tick, period)/2)
		else:
			return int(period/2 + easing.cubic_fast(tick, period)/2)


	slow = cubic_slow
	fast = cubic_fast
	slow_in_fast_out = cubic_slow_in_fast_out

class overlay:
	cache = {}
	tick = 0
	tick2 = 0
	@staticmethod
	def scanlines(surface, density=2, alpha=255):
		overlay.cache['scanlines'] = overlay.cache.get('scanlines', {})
		if not density in overlay.cache['scanlines']:
			s = overlay.cache['scanlines'][density] = \
				pygame.Surface(surface.get_size(), flags=SRCALPHA)
			width, height = surface.get_size()
			for x in xrange(width):
				for y in xrange(height):
					color = (0,0,0) + ((0,) if \
						y % density else (alpha,))
					s.set_at((x, y), color)
		s = overlay.cache['scanlines'][density]
		surface.blit(s, (0, 0))
	
	@staticmethod
	def desaturate(surface, contrast=0, speckle=0, bias=0, color='RGB'):
		"""
		NOTE: loses alpha value, should utilize mask I guess
		"""
		palette = []
		clone = surface.convert(8)
		for i in range(255):
			colors = clone.get_palette_at(i)
			avg = sum(colors) / len(colors) 
			if contrast:
				avg += int(contrast * (127 - abs(127 - avg)))
			if speckle:
				avg += random.randint(-speckle, speckle)
			avg += bias
			avg = min(255, max(0, avg))
			new_color = [0, 0, 0]
			if 'R' in color:
				new_color[0] += avg
			if 'G' in color:
				new_color[1] += avg
			if 'B' in color:
				new_color[2] += avg
			palette.append(tuple(new_color))
		clone.set_palette(palette)
		surface.blit(clone, (0, 0))

	@staticmethod
	def oldtv(surface, cycle=50, color='RGB', contrast_coef=0.1):
		# Get some flickery thingies
		sample_hold_osc = ((overlay.tick % cycle) 
			+ (overlay.tick % cycle / 3)
			- (overlay.tick % cycle * 2)
			- (overlay.tick % 27)) \
			% cycle
		crazy_osc = ((overlay.tick % cycle) 
			+ (overlay.tick % cycle / 3)
			- (overlay.tick % 27)) \
			% cycle
		# First, draw scanlines
		overlay.scanlines(surface, 2, 127)
		overlay.scanlines(surface, 4)
		overlay.scanlines(surface, (overlay.tick % 8 + 3), 127)
		
		# Now, mess with the colors
		overlay.desaturate(surface, contrast=crazy_osc*contrast_coef,
				bias=cycle/50-sample_hold_osc, color=color)
		
		overlay.tick += 1

	@staticmethod
	def strobe(surface, color=(0,0,0), cycle=50):
		TOP, BOTTOM, STEP = 196, 64, 2
		MAX = (TOP-BOTTOM)/STEP
		overlay.cache['strobe'] = overlay.cache.get('strobe', {})
		if not color in overlay.cache['strobe']:
			overlay.cache['strobe'][color] = []
			for x in range(BOTTOM, TOP, STEP):
				s = pygame.Surface(surface.get_size())
				s.fill(color)
				s.set_alpha(x)
				overlay.cache['strobe'][color].append(s)
		tick = (overlay.tick2 % MAX) / 1.0*MAX * 6.28
		frame = int(math.sin(tick)*MAX)
		print frame
		s = overlay.cache['strobe'][color][frame]
		surface.blit(s, (0, 0))
		overlay.tick2 += 1




class Animation(pygame.sprite.Sprite):
	"""
	Animations are auto-animated sprites. Pretty handy for special effects,
	throbbing or spinning things, etc.
	"""
	# TODO generalize to multiple images, ie have it be *effects, as the
	# first effects just generate images
	def __init__(self, image_surface, image_rect, *effects):
		pygame.sprite.Sprite.__init__(self)
		self.source_image = image_surface
		self.source_rect = image_rect
		self.image = image_surface
		self.rect = image_rect
		self.effects = effects
		self.effect_list = EffectList(*effects)
	
	def update(self):
		self.image, self.rect = \
			self.effect_list.do(self.source_image, self.source_rect)
		if self.rect == RECT_UNSET:
			self.kill()

class sprite:


	@staticmethod
	def animate(group, sprite, *a):
		sprite.kill()
		anim = Animation(sprite.image, sprite.rect, *a)
		group.add(anim)


	@staticmethod
	def fly_into(anim_group, sprite, other_sprite, period=100):
		sprite.kill()

		shrink = effects.Grow(period=period, start_size=1.0, end_size=0.0)
		glide = effects.Glide(period=period, start=sprite.rect.topleft,
			end=other_sprite.rect.center, easing=easing.fast, blocking=True)
		anim = Animation(sprite.image, sprite.rect, shrink, glide, effects.Kill())
		anim_group.add(anim)

	@staticmethod
	def float_up_and_vanish(anim_group, sprite, period=100):
		sprite.kill()
		new_r = sprite.rect.copy()
		new_r.top = new_r.top - (sprite.rect.height*1.5)
		grow = effects.Grow(period=period, start_size=1.0, end_size=1.5)
		glide = effects.Glide(period=period, start=sprite.rect.topleft,
			end=new_r.topleft, blocking=True)
		anim = Animation(sprite.image, sprite.rect, grow, glide, effects.Kill())
		anim_group.add(anim)
class comboeffects:
	class Throb(EffectList):
		def __init__(self, period=400, **k):
			EffectList.__init__(self,
					# Then throb
					effects.Loop(
						effects.Grow(period=period,
							start_size=1.0, end_size=0.9,
							easing=easing.fast,
							blocking=True),
						effects.Grow(period=period,
							start_size=0.9, end_size=1.0,
							easing=easing.slow,
							blocking=True)
					),
					**k
			)
			#		effects.Loop(
			#			effects.Spin(period=int(period),
			#				#easing=easing.slow_in_fast_out,
			#				start_angle=-0.5,
			#				end_angle=0.5,
			#				blocking=True),
			#			effects.Spin(period=int(period),
			#				#easing=easing.slow_in_fast_out,
			#				start_angle=0.5,
			#				end_angle=-0.5,
			#				blocking=True),
			#		),
			#		**k
			#)

	class FlipIn(EffectList):
		def __init__(self, time=400, **k):
			period = time / 4
			EffectList.__init__(self,
					# Simultaneously spin and grow
					effects.Spin(period=period), 
					effects.Grow(period=time, start_size=0.0, end_size=1.0,
						after_finished=effects.prune_before,
						easing=easing.slow,
						blocking=True),
						**k
					)
	
	class FlipInAndThrob(EffectList):
		def __init__(self, time=400, throb=400):
			EffectList.__init__(self,
					comboeffects.FlipIn(time, blocking=True),
					#effects.Glide(period=500, start=(0,0), end=(200,200)),
					#effects.FadeOut(period= 100),
					comboeffects.Throb(throb, blocking=True),
				)

def main():
	pygame.init()
	if not pygame.font: print 'Warning, fonts disabled'
	if not pygame.mixer: print 'Warning, sound disabled'
	screen = pygame.display.set_mode((640, 440))
	pygame.display.set_caption('test')
	image, rect = load_image('imagetest.png')
	logo, logo_rect = load_image('logo_1.png')
	
	logo_rect.center = screen.get_rect().center
	anim = Animation(logo, logo_rect, comboeffects.FlipInAndThrob(200))
	anims = pygame.sprite.Group(anim)
	
	background = pygame.Surface(screen.get_size())
	background.fill((200, 200, 200))
	background.blit(image, rect)
	clock = pygame.time.Clock()
	b1 = WaveButton('Test1')
	
	container = AutoTopHalfContainer(
			AutoGridContainer(
				b1,
			),
			proportion=0.8,
		)
	while True:
		for event in pygame.event.get():
			if event.type == QUIT:
				return
			elif event.type == KEYUP:
				if event.key == K_m:
					container[0].append(WaveButton('Test3'))
			
			elif event.type == KEYDOWN or event.type == KEYUP:
				if event.key == K_ESCAPE:
					return
				else:
					b1.on_movement(800)
		screen.blit(background, (0, 0))
		container.update_all()
		container.render(screen)

		anims.update()
		anims.draw(screen)
		
		#overlay.scanlines(screen, 2, 127)
		#overlay.scanlines(screen, 4)
		#overlay.desaturate(screen, 2, 30)
		overlay.oldtv(screen)
		
		pygame.display.flip()
		clock.tick(60)


def change_alpha(surface,alpha=127):
	"""
	WARNING: VERY SLOW.
	"""
	size = surface.get_size()
	try:
		for y in xrange(size[1]):
			for x in xrange(size[0]):
				r,g,b,a = surface.get_at((x,y))
				surface.set_at((x,y),(r,g,b,min(a, int(alpha))))
	except:
		pass
	return surface

def main_dis():
	for i in range(101):
		print i,easing.slow(i, 100),easing.quad_slow(i,100), easing.fast(i, 100), easing.quad_fast(i,100)

if __name__=='__main__':
	main()


