#    CamPlay (C) 2011, released under the GPL 3.0
#    
#    This file is part of the CamPlay Camera Game Library.
#
#    CamPlay is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    CamPlay is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with CamPlay.  If not, see <http://www.gnu.org/licenses/>.


import pygame
import os
import sys
import random
import weakref
import math
import time
from pygame.locals import *
import imagecache 
import ui



__author__ = "Michael Bethencourt"
__copyright__ = "Copyright 2011, Michael Bethencourt"
__license__ = "GPL"
__version__ = "3.0.9"
__email__ = "michaelpb@gmail.com"
__version__ = "0.1.0"


RECT_UNSET = pygame.Rect((-1, -1), (0, 0))

class AnimatedSprite(imagecache.Sprite):
	def __init__(self, *a, **k):
		self.effects = []
		return imagecache.Sprite(self, *a, **k)

	def update(self):
		for effect in self.effects:
			effect.apply(self)
		return


class Seconds(float):
	per_tick = 1.0/60

	def to_ticks(self):
		return int(self * Seconds.per_tick)


class Effect(object):
	def __init__(self, length, easing=None, **k):
		self.length = Seconds(length).to_ticks()
		self.easing = easing
		self.reset()
		self.settings = k

	def reset(self):
		self.tick = 0
	
	def is_done(self):
		return self.tick > self.tick
	
	def apply(self, sprite, tick=None):
		tick = tick if tick else self.tick
		tick = self.easing(tick, self.length) if self.easing else tick
		self.do(sprite, tick)

	def __call__(self, sprite, tick=None):
		return self.apply(sprit, tick)


class EffectList(Effect, list):
	def __init__(self, *a, **k):
		# Init the inside list
		list.__init__(self, a)

		# Use the sum of lengths, if none is specified
		length = k.get('length', [arg.length for arg in a])
		
		# Use linear easing by default
		easing = k.get('easing', None)

		self._original_list = self[:]
		
		Effect.__init__(self, length, easing)
		
		self.reset()

	def reset(self):
		self[:] = self._original_list
		Effect.reset(self)
		
	def is_done(self):
		return len(self) < 1
	
	def do(self, sprite, tick):
		effect = self[0]
		effect.apply(sprite, tick)
		# If finished, pop it off the list
		if effect.is_done():
			self.pop(0)


class Combine(EffectList):
	def do(self, sprite, tick):
		for effect in self[:]:
			effect.apply(sprite, tick)
			if effect.is_done():
				self.remove(effect)

class Loop(EffectList):
	def do(self, sprite, tick):
		effect = self[0]
		effect.apply(sprite, tick)
		# If finished, pop it off the list, and append it to the end
		if effect.is_done():
			self.pop(0)
			self.append(effect)
			effect.reset()



class Blocking(EffectList):
	def __init__(self, blocking_effect, other_effect, **k):
		EffectList.__init__(self, blocking_effect, other_effect, **k)
	
	def is_done(self):
		return self[0].is_done()
	
	def do(self, sprite, tick):
		self[0].apply(sprite, tick)
		if len(self) > 1:
			self[1].apply(sprite, tick)
			if self[1].is_done():
				del self[1]


class Spin(Effect):
	def do(self, sprite, tick):
		max_angle = self.settings.get('angle', 6.28)
		angle = (tick * 1.0 / self.length) * max_angle
		sprite.rotate_ip(angle)

class Grow(Effect):
	def do(self, sprite, tick):
		proportion = self.settings.get('proportion', 2.0)
		if isinstance(proportion, float):
			proportion = (proportion, proportion)
		proportion = (tick*1.0 / self.length) * proportion
		width, height = sprite.prect.size
		sprite.scale((width*proportion[0], height*proportion[0]))


class Glide(Effect):
	def do(self, sprite, tick):
		#location_x = self.settings.get('edge', 'topleft')
		if 'destination' not in self.settings:
			raise ValueError("Destination must be specified.")
		if 'start' not in self.settings:
			self.settings['start'] = sprite.prect.topleft
		distance_left =  self.settings['destination'][0] - self.settings['start'][0]
		distance_top =  self.settings['destination'][1] - self.settings['start'][1]
		new_left = (tick*1.0/self.length) * distance_left + self.settings['destination'][0]
		new_top = (tick*1.0/self.length) * distance_top + self.settings['destination'][1]
		sprite.prect.topleft = (new_left, new_top)


class FadeIn(Effect):
	def do(self, sprite, tick):
		sprite.fade_ip(int((tick*1.0/self.length)*255))

class FadeOut(Effect):
	def do(self, sprite, tick):
		sprite.fade_ip(int((1.0-(tick*1.0/self.length))*255))

# TODO Implement a feature



class easing:
	@staticmethod
	def _pow_easing(tick, period, exp):
		base = math.pow(period, 1.0/exp) * (tick*1.0/period)
		return int(math.pow(base, exp))

	@staticmethod
	def quad_slow(tick, period=100):
		# Quadratic easing
		return easing._pow_easing(tick, period, 2)
		# Is the following faster?
		#base = math.sqrt(period) * (tick*1.0/period)
		#return int(math.pow(base, 2))
	
	@staticmethod
	def cubic_slow(tick, period=100):
		# Cubic easing
		return easing._pow_easing(tick, period, 3)
	
	@staticmethod
	def quad_fast(tick, period=100):
		return period - easing.quad_slow(period-tick, period)
	
	@staticmethod
	def cubic_fast(tick, period=100):
		return period - easing.cubic_slow(period-tick, period)

	@staticmethod
	def cubic_slow_in_fast_out(tick, period=100):
		if tick<period/2:
			return int(easing.cubic_slow(tick, period)/2)
		else:
			return int(period/2 + easing.cubic_fast(tick, period)/2)


	slow = cubic_slow
	fast = cubic_fast
	slow_in_fast_out = cubic_slow_in_fast_out





# Sound management idea: 
#  1. have repeat control (machine gun fire)
#  2. chopping   (siren effect -> intro, middle outro)
class SuperSound(object):
	CHANNEL_COUNT = 8
	last_reserved = 0
	instances = []
	def __init__(self, filepath=None, channels=1, volume=0.8):
		if filepath:
			# Only for extending
			self.sound = imagecache.load_sound(filepath)
		self.channels = [SuperSound.get_channel() for c in xrange(channels)]
		SuperSound.instances.append(weakref.ref(self))
		self.volume = volume
	
	def play(self, **k):
		return self._play(self.sound)
	
	def _play(self, sound, **k):
		found_channel = None
		for channel in self.channels:
			if not channel.get_busy():
				found_channel = channel
				break
		else:
			# Keep the list like an LRU, basically,
			# and stop the oldest one from playing
			found_channel = self.channels.pop(0)
			self.channels.append(found_channel)
			found_channel.stop()
		
		found_channel.play(sound)
		sound.set_volume(self.volume)

		return found_channel



	def update(self):
		pass

	
	@staticmethod
	def updateall():
		for instance in SuperSound.instances:
			if instance():
				instance().update()
		
	@staticmethod
	def get_channel():
		c = SuperSound.last_reserved
		SuperSound.last_reserved = (SuperSound.last_reserved + 1) % SuperSound.CHANNEL_COUNT 
		return pygame.mixer.Channel(c)

class CrossfadeSound(SuperSound):
	NOT_PLAYING, INTRO, MIDDLE, OUTRO = range(4)
	def __init__(self, intro, middle=None, outro=None, fade=1, prefaded=True):
		if not middle and not outro:
			outro = intro + '_outro.ogg'
			middle = intro + '_middle.ogg'
			intro = intro + '_intro.ogg'
		self.intro = imagecache.load_sound(intro)
		self.middle = imagecache.load_sound(middle)
		self.outro = imagecache.load_sound(outro)
		self.elapsed = None
		self.fade_to_middle_at = self.intro.get_length() - fade
		print "FADE TIME", self.fade_to_middle_at
		self.fade_between_middle_at = self.middle.get_length() - fade
		print "FADE TIME", self.fade_between_middle_at
		# Even though we just use 2, 4 is better in case of fast switching
		SuperSound.__init__(self, channels=4)
		self.phase = self.NOT_PLAYING
		self.fade = fade
		self.prefaded = prefaded

		self.intro_channel = None
		self.last_middle_channel = None
	
	def play(self):
		self.phase = self.INTRO
		self.intro_channel = self._play(self.intro)
		self.elapsed = StopWatch()
		#self._play(self.intro, 
		#self.channel.play(self.sound)
	
	def stop(self):
		# TODO revamp all this, relying on the following phases:
		#   (or something similar)
		#      this will keep only one copy of the sound playing
		#      at any given time, is the idea
		#      This is becaause sounds would only be triggered on
		#      edge phases, and even if play() is called twice in
		#      one tick, nothing will go wrong.
		#  NOT_PLAYING
		#  EDGE_INTRO
		#  INTRO
		#  EDGE_INTRO_MIDDLE
		#  INTRO_MIDDLE
		#  EDGE_MIDDLE
		#  MIDDLE
		#  EDGE_MIDDLE_MIDDLE
		#  MIDDLE_MIDDLE
		#  EDGE_MIDDLE_OUTRO
		#  MIDDLE_OUTRO
		#  EDGE_OUTRO
		#  OUTRO
		#  A few annoying corner cases
		#  EDGE_INTRO_OUTRO
		#  INTRO_OUTRO
		#  EDGE_INTRO_MIDDLE_OUTRO
		#  INTRO_MIDDLE_OUTRO
		#  INTRO_OUTRO
		#  
		
		if self.phase not in (self.INTRO, self.MIDDLE):
			return
		self.phase = self.OUTRO
		# Do manual fade out
		#self.ticks = 0
		if self.intro_channel and not self.last_middle_channel:
			self.intro_channel.fadeout(self.fade*1000)
		elif self.last_middle_channel and self.intro_channel:
			# Arbitrary...
			self.last_middle_channel.fadeout(self.fade*500)
			self.intro_channel.fadeout(self.fade*500)
		elif self.last_middle_channel:
			self.last_middle_channel.fadeout(self.fade*1000)
		self.outro_channel = self._play(self.outro)
		self.intro_channel = None
		self.last_middle_channel = None
	
	def update(self):
		if self.phase == self.INTRO:
			# We are playing the intro now.
			# check if we should switch to the middle
			if int(self.elapsed) >= self.fade_to_middle_at:
				if not self.prefaded:
					self.intro_channel.fadeout(self.fade)
				self.last_middle_channel = self._play(self.middle)
				self.phase = self.MIDDLE
				self.elapsed = StopWatch()
				self.intro_channel = None
		elif self.phase == self.MIDDLE:
			if int(self.elapsed) >= self.fade_between_middle_at:
				if not self.prefaded:
					self.last_middle_channel.fadeout(self.fade)
				self.last_middle_channel = self._play(self.middle)
				self.elapsed = StopWatch()


# TODO: eventually add fade in and fade out support
# TODO add sub-second support!	
class StopWatch:
	def __init__(self):
		self.start_time = time.time()

	def __int__(self):
		return int(self.elapsed())
	
	def __float__(self):
		return float(self.elapsed())
	
	def elapsed(self):
		return time.time() - self.start_time 

class PlayList(list):
	instances = []
	def __init__(self, *a, **k):
		self.opts = k
		PlayList.instances.append(weakref.ref(self))
		self.is_playing = False
		list.__init__(self, a)

	# Add 'mood switching', and so on?
	# Maybe in a master "playlist manager"
	def start(self):
		self.playnext()
		self.is_playing = True
	
	def playnext(self):
		song = self.pop(0)
		if self.opts.get('loop', True):
			self.append(song)
		
		path = os.path.join(
			imagecache._sound_cache.datafolder, song)
		if pygame.mixer.music.get_busy():
			pygame.mixer.music.stop()
		pygame.mixer.music.load(path)
		pygame.mixer.music.play()

	def update(self):
		if self.is_playing and pygame.mixer.music.get_busy():
			self.playnext()

	
	@staticmethod
	def updateall():
		for instance in SuperSound.instances:
			if instance():
				instance().update()
	
	def stop(self):
		pygame.mixer.music.fadeout(10)
		self.is_playing = False



def update():
	SuperSound.updateall()
	PlayList.updateall()


def main_old():
	if not pygame.font: print 'Warning, fonts disabled'
	if not pygame.mixer: print 'Warning, sound disabled'
	pygame.init()
	pygame.init()
	screen = pygame.display.set_mode((468, 600))
	pygame.display.set_caption('test')
	image, rect = imagecache.load_image('imagetest.png')
	background = pygame.Surface(screen.get_size())
	background.fill((64, 64, 0))
	clock = pygame.time.Clock()


	imagecache.reset()
	imagecache.reset_sound(datafolder='ih_data')

	angle = 0.1
	gun = SuperSound('fx_gun0.ogg')
	sound = CrossfadeSound('fx_siren00')
	#pygame.mixer.music.load('ih_data/music_0_controlled_decay.ogg')
	#pygame.mixer.music.play()
	playlist = PlayList('music_0_controlled_decay.ogg')
	playlist.start()
	pygame.mixer.music.set_volume(0.5)
	is_held = False
	siren = False
	while True:
		for event in pygame.event.get():
			if event.type == QUIT:
				return
			elif event.type == KEYDOWN and event.key == K_ESCAPE:
				return
			elif event.type == KEYDOWN and event.key == K_m:
				print "PLAYING"
				is_held = True
			elif event.type == KEYUP and event.key == K_m:
				print "STOP PLAYING"
				is_held = False
			elif event.type == KEYUP and event.key == K_s:
				print "STOP PLAYING"
				siren = not siren
				if siren:
					sound.play()
				else:
					sound.stop()
		if is_held:
			gun.play()
		update()
		
		clock.tick(60)





if __name__=='__main__':
	main_old()
